package es.logmeal.sdk;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

public class ARCoreCamsReader {

    public static void main(String[] args) {
        // Usage example
        String testPath = "C:\\Users\\ridou\\OneDrive\\Escritorio";
        List<String> cameraPaths = getIdsPerScene(testPath, ".txt");

        try {
            CameraInfo2 cameraInfo = readARCoreCams(cameraPaths);
            System.out.println("Camera Intrinsics: " + cameraInfo.getCameraIntrinsics());
            System.out.println("Camera-to-World Matrices: " + cameraInfo.getC2WMatrices());
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    private static List<String> getIdsPerScene(String path, String suffix) {
        File[] files = new File(path).listFiles((dir, name) -> name.endsWith(suffix));
        if (files == null) {
            return Collections.emptyList();
        }

        List<String> filePathsSorted = new ArrayList<>();
        Arrays.sort(files);
        for (File file : files) {
            filePathsSorted.add(file.getAbsolutePath());
        }

        return filePathsSorted;
    }

    private static CameraInfo2 readARCoreCams(List<String> cameraPaths) throws IOException {
        CameraInfo2 cameraInfo = new CameraInfo2();

        for (int k = 0; k < cameraPaths.size(); k++) {
            String cameraPath = cameraPaths.get(k);

            try (BufferedReader reader = new BufferedReader(new FileReader(cameraPath))) {
                List<String[]> camTxt = new ArrayList<>();
                String line;
                while ((line = reader.readLine()) != null) {
                    camTxt.add(line.split("\\s+"));
                }

                if (k == 0) {
                    cameraInfo.setCameraIntrinsics(parseCameraIntrinsics(camTxt));
                }

                // EXTRINSIC PARAMETERS
                List<Float> modelMtxList = parseModelMatrix(camTxt.subList(11, 28));
                float[][] c2w = transposeAndReshape(modelMtxList);
                cameraInfo.addC2WMatrix(c2w);
            }
        }

        return cameraInfo;
    }

    private static CameraIntrinsics parseCameraIntrinsics(List<String[]> camTxt) {
        CameraIntrinsics intrinsics = new CameraIntrinsics();
        intrinsics.setWidth(Integer.parseInt(camTxt.get(4)[0].split(":")[1]));
        intrinsics.setHeight(Integer.parseInt(camTxt.get(5)[0].split(":")[1]));
        intrinsics.setFx(Float.parseFloat(camTxt.get(6)[0].split(":")[1]));
        intrinsics.setFy(Float.parseFloat(camTxt.get(7)[0].split(":")[1]));
        intrinsics.setCx(Float.parseFloat(camTxt.get(8)[0].split(":")[1]));
        intrinsics.setCy(Float.parseFloat(camTxt.get(9)[0].split(":")[1]));
        return intrinsics;
    }

    private static List<Float> parseModelMatrix(List<String[]> lines) {
        List<Float> modelMtxList = new ArrayList<>();
        for (String[] line : lines) {
            modelMtxList.add(Float.parseFloat(line[0].split(":")[1]));
        }
        return modelMtxList;
    }

    private static float[][] transposeAndReshape(List<Float> modelMtxList) {
        float[] array = new float[modelMtxList.size()];
        for (int i = 0; i < modelMtxList.size(); i++) {
            array[i] = modelMtxList.get(i);
        }

        float[][] c2w = new float[4][4];
        for (int i = 0; i < 16; i++) {
            c2w[i % 4][i / 4] = array[i];
        }

        return c2w;
    }
}

class CameraInfo2 {
    private CameraIntrinsics cameraIntrinsics;
    private List<float[][]> c2wMatrices;

    public CameraInfo2() {
        this.c2wMatrices = new ArrayList<>();
    }

    public CameraIntrinsics getCameraIntrinsics() {
        return cameraIntrinsics;
    }

    public void setCameraIntrinsics(CameraIntrinsics cameraIntrinsics) {
        this.cameraIntrinsics = cameraIntrinsics;
    }

    public List<float[][]> getC2WMatrices() {
        return c2wMatrices;
    }

    public void addC2WMatrix(float[][] c2w) {
        c2wMatrices.add(c2w);
    }
}

class CameraIntrinsics {
    private int width;
    private int height;
    private float fx;
    private float fy;
    private float cx;
    private float cy;

    // Getters and setters

    public void setWidth(int width) {
        this.width = width;
    }

    public void setHeight(int height) {
        this.height = height;
    }

    public void setFx(float fx) {
        this.fx = fx;
    }

    public void setFy(float fy) {
        this.fy = fy;
    }

    public void setCx(float cx) {
        this.cx = cx;
    }

    public void setCy(float cy) {
        this.cy = cy;
    }
}
