/*
* This class contains all the resources needed to perform Kalman filter
* */
package es.logmeal.sdk;

import org.apache.commons.math3.filter.DefaultMeasurementModel;
import org.apache.commons.math3.filter.DefaultProcessModel;
import org.apache.commons.math3.filter.KalmanFilter;
import org.apache.commons.math3.filter.MeasurementModel;
import org.apache.commons.math3.filter.ProcessModel;
import org.apache.commons.math3.linear.Array2DRowRealMatrix;
import org.apache.commons.math3.linear.ArrayRealVector;
import org.apache.commons.math3.linear.MatrixUtils;
import org.apache.commons.math3.linear.RealMatrix;
import org.apache.commons.math3.linear.RealVector;

import java.util.ArrayList;
import java.util.Arrays;

public class KalmanFilterInfo {
    private float [] poseMatrix;
    private float[]  shift = {0.0F, 0.0F, 0.0F}, difference = {0.0F, 0.0F, 0.0F};
    private KalmanFilter f;
    private float  thr_min = 0.005f, thr_max = 0.05f;
    private int  initialization_counter = 0;

    public KalmanFilterInfo(float[]poseMatrix){
        this.poseMatrix=poseMatrix;
        this.f = initialize_kalman();
    }
    // Given a camera pose and a distance it performs Kalman filter
    public float[] performKalman(float[] copy_poseMatrix, float distance){
        int j;
        if ((distance > thr_min) && (initialization_counter > 10)) {

            for (int i = 0; i < shift.length; i++)
                shift[i] += difference[i];
            j = 0;
            for (int i = 12; i < 15; i++) {
                copy_poseMatrix[i] = copy_poseMatrix[i] - difference[j];
                j++;
            }
        }
        j = 0;
        double[] copy_poseMatrix_xyz = new double[3];
        for (int i = 12; i < 15; i++) {
            copy_poseMatrix_xyz[j] = copy_poseMatrix[i];
            j++;
        }
        f.correct(copy_poseMatrix_xyz);
        j = 0;
        for (int i = 12; i < 15; i++) {
            copy_poseMatrix[i] = (float) copy_poseMatrix_xyz[j];
            j++;
        }
        initialization_counter++;
        return copy_poseMatrix;
    }
    // Computes the distance between the predicted and the current camera position.
    public float getDistance(float[] copy_poseMatrix){
        float  distance = 0;
        int j = 0;
        for (int i = 12; i < 15; i++) {
            copy_poseMatrix[i] = copy_poseMatrix[i] - shift[j];
            j++;
        }
        // Predicts the next position and computes the difference.
        f.predict();
        double[] stateEstimation = f.getStateEstimation();
        ArrayList<Float> x = new ArrayList<>();
        for (int i = 0; i < stateEstimation.length; i += 3)
            x.add((float) stateEstimation[i]);

        j = 0;
        for (int i = 12; i < 15; i++) {
            difference[j] = copy_poseMatrix[i] - x.get(j);
            j++;
        }
        distance = norm(difference);
        return distance;
    }
    // Checks if the distance is smaller than the defined threshold.
    public boolean isDistanceThrMaxConditionSatisfied(float distance){
        if (distance < thr_max)
            return true;
        return false;
    }
    // Computes the norm
    public float norm(float[] data) {
        double aux = 0.0;
        for (int i = 0; i < data.length; i++)
            aux += (data[i] * data[i]);
        return (float) Math.sqrt(aux);
    }
    // It initializes the Kalman filter using default values.
    public KalmanFilter initialize_kalman() {

        float[] cam_xyz = Arrays.copyOfRange(poseMatrix, 12, 15);
        // State initialization (Size X x 1) --- initialState = x in python algorithm
        double[] aux_initialState = {cam_xyz[0], 0, 0, cam_xyz[1], 0, 0, cam_xyz[2], 0, 0};
        RealVector initialState = new ArrayRealVector(aux_initialState);
        double[][] aux_A = {{1, 1, 0.5, 0, 0, 0, 0, 0, 0}, {0, 1, 1, 0, 0, 0, 0, 0, 0}, {0, 0, 1,
                0, 0, 0, 0, 0, 0}, {0, 0, 0, 1, 1, 0.5, 0, 0, 0}, {0, 0, 0, 0, 1, 1, 0, 0, 0}, {0
                , 0, 0, 0, 0, 1, 0, 0, 0}, {0, 0, 0, 0, 0, 0, 1, 1, 0.5}, {0, 0, 0, 0, 0, 0, 0, 1
                , 1}, {0, 0, 0, 0, 0, 0, 0, 0, 1}};
        RealMatrix A = new Array2DRowRealMatrix(aux_A);

        // Measurament function (Size Z x X)
        double[][] aux_H = {{1, 0, 0, 0, 0, 0, 0, 0, 0}, {0, 0, 0, 1, 0, 0, 0, 0, 0}, {0, 0, 0, 0
                , 0, 0, 1, 0, 0}};
        RealMatrix H = new Array2DRowRealMatrix(aux_H);
        // Covariance Matrix is a identity matrix multiplied by 100000 (Size X x X)
        RealMatrix P = MatrixUtils.createRealIdentityMatrix(9);
        P = P.scalarMultiply(1000);

        // Measurement Uncertainty Noise is a identity matrix multiplied by 25 (Size Z x Z)
        RealMatrix R = MatrixUtils.createRealIdentityMatrix(3);
        R = R.scalarMultiply(25);

        // Process Uncertainty Noise (Size X x X)
        double[][] aux_Q = {{25, 0, 0, 0, 0, 0, 0, 0, 0}, {0, 10, 0, 0, 0, 0, 0, 0, 0}, {0, 0, 1,
                0, 0, 0, 0, 0, 0}, {0, 0, 0, 25, 0, 0, 0, 0, 0}, {0, 0, 0, 0, 10, 0, 0, 0, 0}, {0
                , 0, 0, 0, 0, 1, 0, 0, 0}, {0, 0, 0, 0, 0, 0, 25, 0, 0}, {0, 0, 0, 0, 0, 0, 0, 10
                , 0}, {0, 0, 0, 0, 0, 0, 0, 0, 1}};
        RealMatrix Q = new Array2DRowRealMatrix(aux_Q);
        RealMatrix B = null;
        ProcessModel pm = new DefaultProcessModel(A, B, Q, initialState, P);
        MeasurementModel mm = new DefaultMeasurementModel(H, R);
        return new KalmanFilter(pm, mm);
    }
}
