package es.logmeal.sdk;

import android.content.Context;
import android.graphics.Bitmap;
import android.opengl.GLES20;
import android.opengl.GLSurfaceView;
import android.os.Bundle;
import android.os.Environment;
import android.os.Handler;
import android.os.SystemClock;
import android.util.Log;
import android.view.View;
import android.widget.TextView;

import androidx.appcompat.app.AppCompatActivity;

import com.google.ar.core.Anchor;
import com.google.ar.core.Camera;
import com.google.ar.core.Config;
import com.google.ar.core.Frame;
import com.google.ar.core.Session;
import com.google.ar.core.TrackingState;

import com.google.ar.core.exceptions.CameraNotAvailableException;
import com.google.ar.core.exceptions.NotYetAvailableException;
import com.google.ar.core.exceptions.UnavailableApkTooOldException;
import com.google.ar.core.exceptions.UnavailableArcoreNotInstalledException;
import com.google.ar.core.exceptions.UnavailableDeviceNotCompatibleException;
import com.google.ar.core.exceptions.UnavailableSdkTooOldException;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;

import javax.microedition.khronos.egl.EGLConfig;
import javax.microedition.khronos.opengles.GL10;

import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.atomic.AtomicReference;

import es.logmeal.sdk.AngleComputation;
import es.logmeal.sdk.AngleRange;
import es.logmeal.sdk.CameraInfo;
import es.logmeal.sdk.Center;
import es.logmeal.sdk.FolderManagement;
import es.logmeal.sdk.FrameInfo;
import es.logmeal.sdk.GenerateInputAPI;
import es.logmeal.sdk.KalmanFilterInfo;
import es.logmeal.sdk.ManageMatrix;
import es.logmeal.sdk.MyTuple;
import es.logmeal.sdk.SceneCenterComputation;
import es.logmeal.sdk.exceptions.IncompatibleRgbImageType;
import es.logmeal.sdk.helpers.DisplayRotationHelper;
import es.logmeal.sdk.helpers.TrackingStateHelper;
import es.logmeal.sdk.rendering.BackgroundRenderer;

/*This class contains the workflow to calculate all the needed data for the LogMeal API*/

public class LogMealSDK {
    private static final String TAG = LogMealSDK.class.getSimpleName();


    private Path multipleOutPath;
    private final Handler handler = new Handler();
    private MyTuple[] framePerAngle;
    private KalmanFilterInfo kalmanFilterInfo;
    private final AngleRange angleRange = new AngleRange();
    private final ManageMatrix manageMatrix = new ManageMatrix();
    private AngleComputation angleComputation;
    private final SceneCenterComputation sceneCenterComputation = new SceneCenterComputation();
    private ArrayList<Center> centers;
    private FrameInfo frameInfo;
    private FolderManagement folderManagement;
    private ArrayList<float[][]> trajectories_2d;
    private List<double[]> ranges;
    private float[][] initial_focal_2d = null;
    private float[] cam_0_proj, initial_focal = null;
    private double[] relativeTheta;
    private double generalRelativeTheta = 0.0;
    private long startTimeMillis = 0;
    private boolean methodExecuted = false, hasMainFrame = false, started = false,
            isTrajectoryChanged = false;
    private final int numFrames = 100;
    private int numFramesTaken = 0, trajectoriesCounter = 0;

    String description = "", message = "";

    boolean isRecording = true, isCameraInfoNotNull = true, isSuccessful = false;

    String imageFilename = "";
    private String cam_focal_length = "";
    private String cam_principal_point = "";
    private String main_rgb_idx = "";
    private String camera_pose = "";



    public Object[] callLogMealSDK(Session session, Frame frame, Context context){

        try {
            Log.d("isRecording", String.valueOf(isRecording));
            if (isRecording) {

                //Get the current camera pose (is a 1D array).
                //It's important to know that the 12th, 13th and the 14th position of the array
                //are the respective x,y and z of the camera representation.
                Anchor cameraPoseAnchor = session.createAnchor(frame.getCamera().getPose());
                float[] poseMatrix = new float[16];
                cameraPoseAnchor.getPose().toMatrix(poseMatrix, 0);
                //Transformation of the camera pose to a 2D array.
                float[][] aux_poseMatrix_2d = manageMatrix.transformTo2D(poseMatrix);
                float[][] poseMatrix_2d = manageMatrix.transposeMatrix(aux_poseMatrix_2d);
                //The next condition is for the first instance of the execution
                //During the computation of the algorithm the first camera position is needed,
                //so the values is copied into a general variable. Also a method called
                //initializeVariables to initialize important variables is executed.
                if (this.initial_focal == null) {
                    initial_focal = Arrays.copyOf(poseMatrix, poseMatrix.length);
                    initial_focal_2d = new float[poseMatrix_2d.length][poseMatrix_2d[0].length];
                    started = true;
                    for (int i = 0; i < poseMatrix_2d.length; i++) {
                        for (int j = 0; j < poseMatrix_2d[i].length; j++) {
                            initial_focal_2d[i][j] = poseMatrix_2d[i][j];
                        }
                    }
                    Log.d("inicializo", String.valueOf(isRecording));
                    initializeVariables(context);
                    int j = 0;
                    // Save into a 1D array of size 3, the x,y and z of the first camera pose.
                    cam_0_proj = new float[3];
                    for (int i = 12; i < 15; i++) {
                        cam_0_proj[j] = initial_focal[i];
                        j++;
                    }
                }
                // If the needed frames are already taken, finish the process.
                if (numFramesTaken >= (numFrames)) {
                    generateInputForAPI();
                    // el retorn sera: isRecording, started, isTrajectoryChanged
                    return new Object[]{false, started, isTrajectoryChanged, isCameraInfoNotNull, description, message, true};
                }

                // Initialize the Kalman filter.
                if (this.kalmanFilterInfo == null) {
                    kalmanFilterInfo = new KalmanFilterInfo(initial_focal);
                }
                // Given a camera pose, get the distance between the predicted camera pose and
                // the current camera pose and:
                // 1. If the distance is lower than a threshold, perform Kalman to correct the
                //    position.
                // 2. Else, it means that the trajectory has changed and the execution should be
                //    stopped:
                //      2.1 If the main frame image has already been captured and more than 30
                //      frames has been taken, the execution is successful.
                //      2.2 Else, discard the data.
                float[] copy_poseMatrix = Arrays.copyOf(poseMatrix, poseMatrix.length);
                float distance = kalmanFilterInfo.getDistance(copy_poseMatrix);

                if (kalmanFilterInfo.isDistanceThrMaxConditionSatisfied(distance)) {
                    copy_poseMatrix = kalmanFilterInfo.performKalman(copy_poseMatrix, distance);
                } else {
                    isTrajectoryChanged = true;
                    if (hasMainFrame && (numFramesTaken > 30)) {
                        generateInputForAPI();
                        return new Object[]{false,started, isTrajectoryChanged, isCameraInfoNotNull,description,message,true};
                    } else {
                        File file = new File(String.valueOf(multipleOutPath));
                        folderManagement.deleteDirectory(file);
                        clearVariables();
                        return new Object[]{false,started, isTrajectoryChanged, isCameraInfoNotNull,description,message,false};
                    }

                }
                // Get the camera information.
                CameraInfo cameraInfo = frameInfo.getCameraFromFrame(frame, poseMatrix);
                // The generated camera pose with Kalman filter will be used to compute the
                // algorithm.
                float[][] aux_copy_poseMatrix_2d = manageMatrix.transformTo2D(copy_poseMatrix);
                float[][] copy_poseMatrix_2d = manageMatrix.transposeMatrix(aux_copy_poseMatrix_2d);
                computeAlgorithm(copy_poseMatrix_2d, cam_0_proj);
                // Show the the current angle computed and the next angle range
                // This is shown in the "basic" Android UI created for the SDK
                if (cameraInfo != null) {
                    isCameraInfoNotNull = true;
                    description =  multipleOutPath.toString();
                    message =
                                "\n\nAngle: \n" + generalRelativeTheta + "\n[ "
                                        + framePerAngle[numFramesTaken].getAngleRange()[0] + " , "
                                        + framePerAngle[numFramesTaken].getAngleRange()[1] + " ]";

                    }

                // Save the data into the structure data.
                storeDataIntoMyTuple(frame, poseMatrix, cameraInfo);
                cameraPoseAnchor.detach();
            }


        } catch (IOException e) {
            Log.e(TAG, "Exception on the OpenGL thread", e);
        } catch (IncompatibleRgbImageType | NotYetAvailableException e) {
            throw new RuntimeException(e);
        }
    return new Object[]{isRecording,started,isTrajectoryChanged, isCameraInfoNotNull,description,message,isSuccessful};

    }

    public void userStopButton() throws IncompatibleRgbImageType, IOException,
            NotYetAvailableException {
        if (hasMainFrame && (numFramesTaken > 30)) {
            generateInputForAPI();

        } else {
            File file = new File(String.valueOf(multipleOutPath));
            folderManagement.deleteDirectory(file);
            clearVariables();
        }

    }



    // Method used to set the path.
    public void setMultipleOutPath(Long path) {
        multipleOutPath =
                Paths.get(Environment.getExternalStoragePublicDirectory(Environment.DIRECTORY_DOCUMENTS).getAbsolutePath(), "logmeal", path.toString());
        if (!Files.exists(multipleOutPath)) {
            try {
                Files.createDirectory(multipleOutPath);
                Log.d("Directorio", "Directorio creado");
            } catch (IOException e) {
                Log.e(TAG, "Error creating output directory", e);
            }
        }
    }

    // Method to initialize all the needed variables.
    public void initializeVariables(Context context) {
        // Generate the range of the angles where the frames will be taken.
        centers = new ArrayList<>();
        trajectories_2d = new ArrayList<>();
        ranges = angleRange.getRanges();
        // The ranges are added to the data structure that contains the needed data to generate the
        // input for the API.
        framePerAngle = new MyTuple[numFrames];
        for (int i = 0; i < numFrames; i++) {
            framePerAngle[i] = new MyTuple(ranges.get(i), false);
        }
        frameInfo = new FrameInfo(multipleOutPath, context);
        folderManagement = new FolderManagement();
        relativeTheta = new double[numFrames];
        //
        centers.add(trajectoriesCounter, sceneCenterComputation.initialize_scene_center());
        trajectories_2d.add(trajectoriesCounter, initial_focal_2d);
        trajectoriesCounter++;

    }

    // Method to clear the used variables (it may not be needed).
    public void clearVariables() {
        kalmanFilterInfo = null;
        angleComputation = null;
        frameInfo = null;
        initial_focal = null;
        initial_focal_2d = null;
        numFramesTaken = 0;
        trajectoriesCounter = 0;
        //ranges = null;
        generalRelativeTheta = 0;
        description = "";
        message = "";
        isRecording = true;
        isCameraInfoNotNull = true;
        //centers = null;
        //framePerAngle = null;
        //relativeTheta = null;
        //trajectories_2d = null;
        methodExecuted = false;
        hasMainFrame = false;
        started = false;
        isTrajectoryChanged = false;
        cam_0_proj = null;
        folderManagement = null;

    }

    // Next method compute the logic to execute all the algorithms to generate the raw data
    // that will be needed such as: centers, centers with weight, angles and trajectories.
    public void computeAlgorithm(float[][] poseMatrix_2d, float[] cam_0_proj) {
        // Update the center using all the camera poses of the trajectory, the current camera pose,
        // and the previous center.
        centers.add(trajectoriesCounter,
                sceneCenterComputation.update_scene_center_from_cams(trajectories_2d,
                        trajectories_2d.get(trajectoriesCounter - 1)));
        // Apply weight to the center that has been already updated.
        float[] curr_center =
                sceneCenterComputation.apply_weight_scene_center();
        // Update the y of the first camera pose replacing it by the y of the weighted center.
        cam_0_proj[1] = curr_center[1];
        // Compute the angle based on the center, first camera and all the camera poses.
        angleComputation = new AngleComputation(cam_0_proj, curr_center, trajectories_2d);
        relativeTheta = angleComputation.getAngleWithRespectToFirstCam();
        // General variable to get the angle of the current iteration.
        generalRelativeTheta = relativeTheta[relativeTheta.length - 1];
        // The current camera pose is added to the trajectory.
        trajectories_2d.add(trajectoriesCounter, poseMatrix_2d);
        trajectoriesCounter++;
    }

    // This method generates the input for the API: the sequence of video, the JSON and the main
    // RGB image.
    public void generateInputForAPI() throws IOException, IncompatibleRgbImageType,
            NotYetAvailableException {
        // Generate the video and JSON
        description = "Generating INPUT for API";
        message = "Creating video sequence...";
        GenerateInputAPI generateInputAPI = new GenerateInputAPI(multipleOutPath, framePerAngle);
        generateInputAPI.generateInputsForAPI();
        camera_pose = generateInputAPI.getCamera_pose();
        cam_focal_length = generateInputAPI.getCam_focal_length();
        cam_principal_point = generateInputAPI.getCam_principal_point();
        main_rgb_idx = generateInputAPI.getMain_rgb_idx();
        // Deletes all the taken images except the main RGB image.
        int closestIndex = generateInputAPI.getClosestIndex();
        imageFilename = "image"+closestIndex+".jpeg";

        File file = new File(String.valueOf(multipleOutPath));
        FolderManagement.deleteDirectoryExcept(file, closestIndex);

        //Resize image
        ImageProcessor.resizeAndSaveImage(Paths.get(multipleOutPath.toString(),imageFilename).toString(), Paths.get(multipleOutPath.toString(),"image.jpeg").toString(), 480, 360);

        clearVariables();
    }


    // This method checks if the current angle is between the current range angle to save the
    // needed data into the data structure MyTuple.

    // The data saved are: the RGB image of the current frame, the angle, the path of the folder,
    // the current camera pose, the focal length and principal point of the camera and a boolean
    // to warn that the current range already has an angle.
    public void storeDataIntoMyTuple(Frame frame, float[] poseMatrix, CameraInfo cameraInfo) throws IncompatibleRgbImageType, IOException, NotYetAvailableException {
        // If it is the first time that a data is going to be stored into the data structure,
        // a delay of 2 sec it's added because the values of the angle and center are unstable.
        if (!methodExecuted) {
            long currentTimeMillis = SystemClock.uptimeMillis();

            if (startTimeMillis == 0) {
                startTimeMillis = currentTimeMillis;
                handler.postDelayed(new Runnable() {
                    @Override
                    public void run() {
                        if (!(framePerAngle[numFramesTaken].getHasValue())) {
                            if ((framePerAngle[numFramesTaken].getAngleRange()[0] <= generalRelativeTheta) && (framePerAngle[numFramesTaken].getAngleRange()[1] > generalRelativeTheta)) {
                                try {
                                    frameInfo.getImageData(frame, numFramesTaken);
                                } catch (NotYetAvailableException |
                                         IncompatibleRgbImageType | IOException e) {
                                    throw new RuntimeException(e);
                                }
                                framePerAngle[numFramesTaken].setAngle(generalRelativeTheta);
                                framePerAngle[numFramesTaken].setRgbUri(multipleOutPath.toString());
                                framePerAngle[numFramesTaken].setCameraPose(poseMatrix);
                                framePerAngle[numFramesTaken].setFocalLength(cameraInfo.getFocalLength());
                                framePerAngle[numFramesTaken].setPrincipalPoint(cameraInfo.getPrincipalPoint());
                                framePerAngle[numFramesTaken].setHasValue(true);
                                numFramesTaken += 1;
                            }
                        }
                        methodExecuted = true;
                    }
                }, 2000);
            }
        } else {
            if (!(framePerAngle[numFramesTaken].getHasValue())) {
                if ((framePerAngle[numFramesTaken].getAngleRange()[0] <= generalRelativeTheta) && (framePerAngle[numFramesTaken].getAngleRange()[1] > generalRelativeTheta)) {
                    int epsilon = 5;

                    if ((90 - epsilon < generalRelativeTheta) && (generalRelativeTheta < 90 + epsilon) && !hasMainFrame) {
                        Log.d("GeneralRelativeTheta", String.valueOf(generalRelativeTheta));
                        hasMainFrame = true;
                    }
                    Log.d("ValueGeneral", generalRelativeTheta + "---" + hasMainFrame +
                            " range: [" + framePerAngle[numFramesTaken].getAngleRange()[0] + ", " + framePerAngle[numFramesTaken].getAngleRange()[1] + "]");
                    frameInfo.getImageData(frame, numFramesTaken);
                    framePerAngle[numFramesTaken].setAngle(generalRelativeTheta);
                    framePerAngle[numFramesTaken].setRgbUri(multipleOutPath.toString());
                    framePerAngle[numFramesTaken].setCameraPose(poseMatrix);
                    framePerAngle[numFramesTaken].setFocalLength(cameraInfo.getFocalLength());
                    framePerAngle[numFramesTaken].setPrincipalPoint(cameraInfo.getPrincipalPoint());
                    framePerAngle[numFramesTaken].setHasValue(true);
                    numFramesTaken += 1;
                }
            }
        }
    }
    public String getImageFilename(){
        return imageFilename;
    }
    public String getPath(){
        return multipleOutPath.toString();
    }
    public String getImagePath(){
        return Paths.get(this.getPath(),"image.jpeg").toString();
    }
    public String getSequencePath(){
        return Paths.get(this.getPath(),"output.avi").toString();
    }
    public String getCam_focal_length(){
        return cam_focal_length;
    }
    public String getCam_principal_point(){
        return cam_principal_point;
    }
    public String getMain_rgb_idx(){
        return main_rgb_idx;
    }
    public String getCamera_pose(){
        return camera_pose;
    }
}
