# LogMeal SDK
This is the Android SDK of LogMeal. It's main purpose is to aid in the generation of needed files to use Depth Endpoints.
It can be used as a low level SDK withing ArCore or as a camera widget that obtains the needed depth data.

## 1. How to add this library to your project
1. Download LogMealSDK-release.aar from releases
2. Inside your Android Studio project select File > Project Structure
3. Select dependencies in left sidebar
4. Select your app under Modules (middle sidebar)
5. Under ***Declared Dependencies*** click the plus button "+"
6. Select ***JAR/ARR Dependency***
7. Find the downloaded AAR file and choose implementation
8. Done

## 2. Workflow

This SDK provides a capture mode that allows obtaining data that can be used for the API.
The SDK starts a capture process and will inform you when it is finished. Once finished, it generates all the needed data to call the API.

## 3. How to use
Typically an ArCore application has an OnDrawFrame method where it initialize all the needed tools. 
The SDK is integrated into this flow to obtain the necessary data. It is assumed that you will implement
an application where the user actives the recording mode so that the process can start.
### 3.1 Add atribute to class
Define a LogMealSDK attribute:
```java  
LogMealSDK logMealSDK = new LogMealSDK();
 ```

### 3.2  Add record on `onDrawFrame`
Inside
Inside onDrawFrame. Using the LogMealSDK class we can capture the information. The method called
callLogMealSDK returns an Object array that contains the needed data to control the workflow.
The attribute isRecording checks if it is recording yet or not.
The attribute started tells if the process of capture mode has started or not.
The attribute isTrajectoryChanged checks if the trajectory has changed or not.
The attribute isCameraInfoNotNull is needed to check if the information stored in description and
message are not null.
The 
```java  
	//Check if user has activated recording mode:
    if (currentState.get() == AppState.RECORDING) {
	    //Check if SDK is still recording:
        if(isRecording){
            Object[] response =  logMealSDK.callLogMealSDK(session,frame,this);
            Log.d("Hola", "Adios");
            isRecording = (boolean) response[0];
            started = (boolean) response[1];
            isTrajectoryChanged = (boolean) response[2];
            isCameraInfoNotNull = (boolean) response[3];
            description = (String) response[4];
            message = (String) response[5];
        }
    }
 ```
### 3.3  Obtain result
After the previous call, you must verify if the SDK has finished capturing the data
```java  
       if(!isRecording){
			if(isSuccessful){
               //The information needed for the api is obtained by doing:
                String mainImagePath = logMealSDK.getImagePath()
                String sequencePath = logMealSDK.getSequencePath()
                String camFocalLength = logMealSDK.getCam_focal_length();
                String camPrincipalPoint = logMealSDK.getCam_principal_point();
                String mainRgbIdx = logMealSDK.getMain_rgb_idx();
                String cameraPose = logMealSDK.getCamera_pose();
                //With these data call the API
			}
			else{
			    //Manage Errors
			}
        currentState.set(AppState.IDLE);
        return;
        }
 ```
### 3.4 Add posibility to stop recording
In the case that the user is given the possibility of cutting the capture, userStopButton() must be used:
```java  
        if (currentState.get() == AppState.IDLE && started && !isTrajectoryChanged) {
            logMealSDK.userStopButton();
            currentState.set(AppState.IDLE);
            started = false;
        }
 ```
## 3. Important classes
### 3.1 LogMealSDK
It's purpose is to hold all the data from the captured frame, this data should be used with the LogMeal API. It has the detailed data in the following methods:
```java  
    public Object[] callLogMealSDK(Session session, Frame frame, Context context);
    public void userStopButton() throws IncompatibleRgbImageType, IOException, NotYetAvailableException
    public String getImageFilename()
    public String getPath()
    public String getCam_focal_length()
    public String getCam_principal_point()
    public String getMain_rgb_idx()
    public String getCamera_pose()

 ```

LogMeal API's Depth Endpoint requires the following data:
- Field=image : Path to image file: `logMealSDK.getImagePath()`
- Field=sequence : Path to sequence file: `logMealSDK.getSequencePath()`
- Field=main_rgb_idx : Index of main RGB image: `logMealSDK.getMain_rgb_idx()`
- Field=cam_focal_length : Depth Camera focal length  `logMealSDK.getCam_focal_length()`
- Field=cam_principal_point : Depth Camera principal point `logMealSDK.getCam_principal_point()`
- Field=camera_pose : Depth Camera poses `logMealSDK.getCamera_pose()`
- Field=depth_version: `android | ios`

### 3.6 callLogMealSDK
It is used to calculate all the needed information that will be retrieved using the getters. It 
returns an Object[] containing all the needed data to create an easy workflow. 

### 3.6 userStopButton
It is used so that the user can manually finish the capture process.


### 3.6 ArCoreUtils
It's purpose is to simplify and centralize details about device support. It has one static method that returns an instance of `ArCoreSupport`. If ARCore wasn't installed, it will automatically show a popup to install it.
```java 
    public static ArCoreSupport isDeviceSupported(Context context)
```

### 3.7 ArCoreSupport
It contains the support information for the device. It has three getters:
```java 
    public boolean isArCoreSupported() // ARCore is supported in this device?
    public boolean isDepthSupported() // ARCore's Depth API is supported in this device?
    public ArCoreApk.Availability getArCoreAvailability() // ARCore's availability for this device. Checkout https://developers.google.com/ar/reference/java/com/google/ar/core/ArCoreApk.Availability
```


### 3.8 PermissionsHelper
It's a simply helper to ask for camera permission.
```java
    // Check to see we have the necessary permissions for this app.
    public static boolean hasPermissions(Activity activity)
    // Check to see we have the necessary permissions for this app, and ask for them if we don't.
    public static void requestPermissions(Activity activity)
    // Launch Application Setting to grant permission.
    public static void launchPermissionSettings(Activity activity)
    // Check to see if we need to show the rationale for this permission.
    public static boolean shouldShowRequestPermissionRationale(Activity activity)
```
The helper is usually used like this (in a activity)
```java
    @Override
    protected void onResume() {
        // ...

        // ARCore requires camera permission to operate.
        if (!PermissionsHelper.hasPermissions(this)) {
        PermissionsHelper.requestPermissions(this);
        }
    }

    @Override
    public void onRequestPermissionsResult(int requestCode, String[] permissions, int[] results) {
        super.onRequestPermissionsResult(requestCode, permissions, results);
        if (!PermissionsHelper.hasPermissions(this)) {
            Toast.makeText(this, "Camera and write storage permissions are needed to run this application", Toast.LENGTH_LONG)
                    .show();
            if (!PermissionsHelper.shouldShowRequestPermissionRationale(this)) {
                PermissionsHelper.launchPermissionSettings(this);
            }
            finish();
        }
    }
```
