/*
 * Copyright 2021 Google LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package es.logmeal.sdk.sample;

import android.opengl.GLES20;
import android.opengl.GLSurfaceView;
import android.os.Bundle;
import android.os.Environment;
import android.os.Handler;
import android.util.Log;
import android.view.View;
import android.widget.TextView;

import androidx.appcompat.app.AppCompatActivity;

import com.google.android.material.textfield.TextInputEditText;
import com.google.ar.core.Camera;
import com.google.ar.core.Config;
import com.google.ar.core.Frame;
import com.google.ar.core.Session;
import com.google.ar.core.TrackingState;

import com.google.ar.core.exceptions.CameraNotAvailableException;
import com.google.ar.core.exceptions.NotYetAvailableException;
import com.google.ar.core.exceptions.UnavailableApkTooOldException;
import com.google.ar.core.exceptions.UnavailableArcoreNotInstalledException;
import com.google.ar.core.exceptions.UnavailableDeviceNotCompatibleException;
import com.google.ar.core.exceptions.UnavailableSdkTooOldException;

import java.io.IOException;

import javax.microedition.khronos.egl.EGLConfig;
import javax.microedition.khronos.opengles.GL10;

import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.atomic.AtomicReference;

import es.logmeal.sdk.AngleComputation;
import es.logmeal.sdk.AngleRange;
import es.logmeal.sdk.Center;
import es.logmeal.sdk.FolderManagement;
import es.logmeal.sdk.FrameInfo;
import es.logmeal.sdk.KalmanFilterInfo;
import es.logmeal.sdk.LogMealSDK;
import es.logmeal.sdk.ManageMatrix;
import es.logmeal.sdk.MyTuple;
import es.logmeal.sdk.SceneCenterComputation;
import es.logmeal.sdk.SendDepthRequest;
import es.logmeal.sdk.exceptions.IncompatibleRgbImageType;
import es.logmeal.sdk.helpers.DisplayRotationHelper;
import es.logmeal.sdk.helpers.TrackingStateHelper;
import es.logmeal.sdk.rendering.BackgroundRenderer;

/**
 * This is a simple example that shows how to use ARCore Raw Depth API. The application will display
 * a 3D point cloud and allow the user control the number of points based on depth confidence.
 */
public class AdvancedIntegrationExampleActivity extends AppCompatActivity implements GLSurfaceView.Renderer {
    private static final String TAG = AdvancedIntegrationExampleActivity.class.getSimpleName();

    private enum AppState {IDLE, PICTURE, RECORDING}
    private final Object frameInUseLock = new Object();
    private final BackgroundRenderer backgroundRenderer = new BackgroundRenderer();
    private final TrackingStateHelper trackingStateHelper = new TrackingStateHelper(this);
    private final AtomicReference<AppState> currentState = new AtomicReference<>(AppState.IDLE);
    private Session session;
    private DisplayRotationHelper displayRotationHelper;
    private GLSurfaceView surfaceView;
    private TextView outPathTextView, resultTextView;

    private boolean  started = false, isTrajectoryChanged = false;
    LogMealSDK logMealSDK = new LogMealSDK();
    String description = "", message = "";

    boolean isRecording = true, isCameraInfoNotNull = true, isSuccessful = false;

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        Path outPath;
        super.onCreate(savedInstanceState);

        setContentView(R.layout.activity_advanced_integration_example);
        surfaceView = findViewById(R.id.surfaceView);
        outPathTextView = findViewById(R.id.textView4);
        resultTextView = findViewById(R.id.advancedResultTextView);

        displayRotationHelper = new DisplayRotationHelper(/*context=*/ this);

        // Set up rendering.
        surfaceView.setPreserveEGLContextOnPause(true);
        surfaceView.setEGLContextClientVersion(2);
        surfaceView.setEGLConfigChooser(8, 8, 8, 0, 16, 0);
        surfaceView.setRenderer(this);
        surfaceView.setRenderMode(GLSurfaceView.RENDERMODE_CONTINUOUSLY);
        surfaceView.setWillNotDraw(false);

        outPath =
                Paths.get(Environment.getExternalStoragePublicDirectory(Environment.DIRECTORY_DOCUMENTS).getAbsolutePath(), "logmeal");

        if (!Files.exists(outPath)) {
            try {
                Files.createDirectory(outPath);
            } catch (IOException e) {
                Log.e(TAG, "Error creating output directory", e);
            }
        }
    }

    @Override
    protected void onResume() {
        super.onResume();

        try {
            if (session == null) session = new Session(this);

        } catch (UnavailableDeviceNotCompatibleException | UnavailableSdkTooOldException |
                 UnavailableApkTooOldException | UnavailableArcoreNotInstalledException e) {
            Log.e(TAG, "Exception creating session", e);
        }

        try {
            synchronized (frameInUseLock) {
                // Enable raw depth estimation and auto focus mode while ARCore is running.
                Config config = session.getConfig();
                config.setFocusMode(Config.FocusMode.AUTO);
                config.setDepthMode(Config.DepthMode.AUTOMATIC);
                session.configure(config);
                session.resume();
            }
        } catch (CameraNotAvailableException e) {
            session = null;
            return;
        }
        surfaceView.onResume();
        displayRotationHelper.onResume();
    }

    @Override
    public void onPause() {
        super.onPause();
        if (session != null) {
            // Note that the order matters - see note in onResume().
            // GLSurfaceView is paused before pausing the ARCore session, to prevent onDrawFrame
            // () from
            // calling session.update() on a paused session.
            displayRotationHelper.onPause();
            surfaceView.onPause();
            if (currentState.get() == AppState.PICTURE) currentState.set(AppState.IDLE);
            session.pause();
        }
    }

    @Override
    protected void onDestroy() {
        if (session != null) {
            session.close();
            session = null;
        }
        super.onDestroy();
    }

    @Override
    public void onSurfaceCreated(GL10 gl, EGLConfig config) {
        GLES20.glClearColor(0.1f, 0.1f, 0.1f, 1.0f);
        // Prepare the rendering objects. This involves reading shaders, so may throw an
        // IOException.
        try {
            backgroundRenderer.createOnGlThread(/*context=*/ this);
        } catch (IOException e) {
            Log.e(TAG, "Failed to read an asset file", e);
        }
    }

    @Override
    public void onSurfaceChanged(GL10 gl, int width, int height) {
        displayRotationHelper.onSurfaceChanged(width, height);
        GLES20.glViewport(0, 0, width, height);
    }

    @Override
    public void onDrawFrame(GL10 gl) {
        // Clear screen to notify driver it should not load any pixels from previous frame.
        GLES20.glClear(GLES20.GL_COLOR_BUFFER_BIT | GLES20.GL_DEPTH_BUFFER_BIT);
        // Do not render anything or call session methods until session is created.
        if (session == null) return;
        displayRotationHelper.updateSessionIfNeeded(session);
        try {
            session.setCameraTextureName(backgroundRenderer.getTextureId());
            Frame frame = session.update();
            Camera camera = frame.getCamera();
            backgroundRenderer.draw(frame);
            trackingStateHelper.updateKeepScreenOnFlag(camera.getTrackingState());
            if (camera.getTrackingState() != TrackingState.TRACKING) return;

            if (currentState.get() == AppState.PICTURE) {
                currentState.set(AppState.IDLE);
            }

            if (currentState.get() == AppState.RECORDING) {
                if(isRecording){
                    // Call the SDK to calculate the information
                    Object[] response =  logMealSDK.callLogMealSDK(session,frame,this);
                    isRecording = (boolean) response[0];
                    started = (boolean) response[1];
                    isTrajectoryChanged = (boolean) response[2];
                    isCameraInfoNotNull = (boolean) response[3];
                    description = (String) response[4];
                    message = (String) response[5];
                    isSuccessful = (boolean) response[6];
                }

                if (isCameraInfoNotNull) {
                    runOnUiThread(() -> {
                        outPathTextView.setText(description);
                        resultTextView.setText(message);
                    });
                }

                if(!isRecording){
                    currentState.set(AppState.IDLE);
                    return;
                }

            }
            if (currentState.get() == AppState.IDLE && started && !isTrajectoryChanged) {
                logMealSDK.userStopButton();
                currentState.set(AppState.IDLE);
                started = false;

                runOnUiThread(() -> {
                    outPathTextView.setText("Files deleted successfully\n");
                    resultTextView.setText("\nThe capture failed.");
                });

            }


        } catch (CameraNotAvailableException | IOException e) {
            Log.e(TAG, "Exception on the OpenGL thread", e);
        } catch (IncompatibleRgbImageType | NotYetAvailableException e) {
            throw new RuntimeException(e);
        }
    }

    public void captureDepthSnapshot(View view) {
        currentState.set(AppState.PICTURE);
    }

    // Method that manages the state of the execution.
    public void captureDepthMultipleSnapshot(View view) {
        if (currentState.get() == AppState.RECORDING) {
            currentState.set(AppState.IDLE);

        } else {
            currentState.set(AppState.RECORDING);
            logMealSDK.setMultipleOutPath(System.currentTimeMillis());
            isRecording = true;
        }
    }
    public void onSendDepthRequestClick(View view) {
        // Retrieve the information from the SDK and call the API
        TextInputEditText textInputEditText = findViewById(R.id.textInput);
        String authToken = textInputEditText.getText().toString();
        String fileName = logMealSDK.getImageFilename();
        String imagePath = logMealSDK.getImagePath();
        String sequencePath = logMealSDK.getSequencePath();

        String camFocalLength = logMealSDK.getCam_focal_length();
        String camPrincipalPoint = logMealSDK.getCam_principal_point();
        String mainRgbIdx = logMealSDK.getMain_rgb_idx();
        String cameraPose = logMealSDK.getCamera_pose();
        String version = "android";

        // Create an instance of SendDepthRequest
        SendDepthRequest sendDepthRequest = new SendDepthRequest(authToken, imagePath, fileName,sequencePath,
                camFocalLength, camPrincipalPoint, mainRgbIdx, cameraPose, version);

        // Create an ExecutorService
        ExecutorService executorService = Executors.newSingleThreadExecutor();

        // Submit the task and get a Future representing the pending completion of the task
        Future<Void> future = executorService.submit(sendDepthRequest);



        // You can use future.get() if you want to wait for the task to complete before continuing
        // creo que sería interesante mostrar simplemente si el request fue exitosos o no

        // Don't forget to shutdown the executor service when you're done
        executorService.shutdown();
    }




}
